/*
* QuickFox Notes by Amin E (InBasic)
* eMail: inb.cor@gmail.com
*
* Source Code License: Mozilla Public License, version 1.1
*/
QuickFoxNotes.io = {
	ioTitle: 0,
	ioDescription: 1,
	
	accessType: null,	//0 is Bookmarks, 1 is local file

	initialize: function () {
		this.accessType = QuickFoxNotes.prefs.getIntPref("repositoryType");
		var confirm = -1;
		while (confirm == -1) 
			confirm = this[this.accessType ? "sqlite" : "bookmark"].initialize();
			//confirm = this[this.accessType ? "local" : "bookmark"].initialize();

		if (!confirm)
			confirm = this.bookmark.initialize();

	},
	getList: function () {
		return this.accessType ? this.sqlite.getIndices() : this.bookmark.getBookmarkIDs();
		//return this.accessType ? this.local.getFilesIndex() : this.bookmark.getBookmarkIDs();
	},
	read: function (type, ID) {
		return this[this.accessType ? "sqlite" : "bookmark"].read(type, ID);
		//return this[this.accessType ? "local" : "bookmark"].read(type, ID);
	},
	write: function (type, ID, data) {
		return this[this.accessType ? "sqlite" : "bookmark"].write(type, ID, data);
		//return this[this.accessType ? "local" : "bookmark"].write(type, ID, data);
	},
	getItemIndex: function (ID) {
		return this[this.accessType ? "sqlite" : "bookmark"].getItemIndex(ID);
		//return this[this.accessType ? "local" : "bookmark"].getItemIndex(ID);
	},
	moveItem: function (ID, newLocation) {
		return this[this.accessType ? "sqlite" : "bookmark"].moveItem(ID, newLocation);
		//return this[this.accessType ? "local" : "bookmark"].moveItem(ID, newLocation);
	},
	remove: function (IDs) {
		return this[this.accessType ? "sqlite" : "bookmark"].remove(IDs);
		//return this[this.accessType ? "local" : "bookmark"].remove(IDs);
	},
	release: function () {	//This function will release the database without need to re-initiating it again! (So Dropbox can replace file although Firefox is open)
		if (this.accessType == 1)
			this.sqlite.releaseDatabase();
	},
	//Soft switch to local mode
	softSwitch: {
		toLocal: function() {
			QuickFoxNotes.prefs.setCharPref("lFolderName", "");	//Reset Directory
			QuickFoxNotes.prefs.setCharPref("sqlitePath", "");	//Reset sqlite data storage
			QuickFoxNotes.prefs.setIntPref("repositoryType", 1);
			var confirm = -1;
			while (confirm == -1)
				confirm = QuickFoxNotes.io.sqlite.initialize();	//If true continue; if false, stop; if -1 retry
				//confirm = QuickFoxNotes.io.local.initialize();	//If true continue; if false, stop; if -1 retry
				
			alert(QuickFoxNotes.stringsBundle.getString(confirm ? 'io12' : 'io13'));
		},
		toBookmark: function () {
			QuickFoxNotes.prefs.setIntPref("repositoryType", 0);
			alert(QuickFoxNotes.stringsBundle.getString('io12'));
		}
	},
	//Call this method to force QFN clear old repository and mark all notes as new!
	ioSwitch: function () {
		try{
			this.bookmark.initialize();
			var confirm = -1;
			while (confirm == -1)
				confirm = this.sqlite.initialize();	//If true continue; if false, stop; if -1 retry
				//confirm = this.local.initialize();	//If true continue; if false, stop; if -1 retry
			if (!confirm)
				return;
				
			if (window.confirm(QuickFoxNotes.stringsBundle.getString('io3') + " \"" + 
						  (this.accessType == 0 ? QuickFoxNotes.stringsBundle.getString('io11') : QuickFoxNotes.stringsBundle.getString('io10')) +"\"\n" +
						   QuickFoxNotes.stringsBundle.getString('io4') + "\n\n" +
						   QuickFoxNotes.stringsBundle.getString('io5') + " " +
						   QuickFoxNotes.stringsBundle.getString('io6') + "\n\n" +
						   QuickFoxNotes.stringsBundle.getString('io7'))) {
				//Clear old data
				if (this.accessType == 0) {	//New repository would be local
					this.sqlite.remove(this.sqlite.getIndices());
					//Clear local folder
					//this.local.remove(this.local.getFilesIndex());
					//this.local.files = [];
				}
				else {	//New repository would be bookmark
					this.bookmark.remove(this.bookmark.getBookmarkIDs());
				}
				//Mark all notes as not saved
				var tab = QuickFoxNotes.tabs.getElementsByTagName("tab");
				for (var i = 1; i < QuickFoxNotes.currentNumberOfPanels; i++) {	//First note is dummy
					tab[i].saveMe = true;
					QuickFoxNotes.getEditor(i).saveMe = true;
					tab[i].bookmarkId = null;
				}
				var menupopup = _("editor.toolbar6.menupopup");
				for (var i = 0 ; i < menupopup.childNodes.length ; i++) {
					menupopup.childNodes[i].saveMe = true;
					menupopup.childNodes[i].bookmarkId = null;
				}
				QuickFoxNotes.prefs.setIntPref("repositoryType", this.accessType ? 0 : 1);
				QuickFoxNotes.io.initialize.call(QuickFoxNotes.io);	
				
				QuickFoxNotes.savePanels();
				QuickFoxNotes.renderTabs();	//The new repository has a different title!
				alert(QuickFoxNotes.stringsBundle.getString('io8'));
			}
		} catch (e) {
			alert(QuickFoxNotes.stringsBundle.getString('io9') + "\n\n" + e.lineNumber + " : "+ e.message);
		}
	},
	/**Change the bookmarks folder*/
	changeBookmarksFolder: function () {
		var check = {value: false};
		var input = {value: QuickFoxNotes.prefs.getComplexValue("FolderName", Components.interfaces.nsISupportsString).data};
		var result = QuickFoxNotes.prompts.prompt(null, QuickFoxNotes.stringsBundle.getString('changeRepositoryTitle'), QuickFoxNotes.stringsBundle.getString('changeRepository'), input, null, check);
	
		if (result && input.value && !(/\/*$/.exec(input.value)[0]) ) {	//input.value should be non empty and also not terminate with '/'
			var str = Components.classes["@mozilla.org/supports-string;1"]
				.createInstance(Components.interfaces.nsISupportsString);
				str.data = input.value;
			QuickFoxNotes.prefs.setComplexValue("FolderName", Components.interfaces.nsISupportsString, str); 
			alert(QuickFoxNotes.stringsBundle.getString('repositorySuccess'));
		}
		else if (result)
			alert(QuickFoxNotes.stringsBundle.getString('repositoryWarning'));
	}
}
/**Bookmarks I/O service*/
QuickFoxNotes.io.bookmark = {
	bmsvc: null,
	annotationService: null,
	
	rootID: null,

	initialize: function () {
		this.bmsvc = Components.classes["@mozilla.org/browser/nav-bookmarks-service;1"]
			.getService(Components.interfaces.nsINavBookmarksService);
		this.annotationService = Components.classes["@mozilla.org/browser/annotation-service;1"]
			.getService(Components.interfaces.nsIAnnotationService),
			
		this.rootID = this.locateRoot();
		return true;
	},
	locateRoot: function () {
		function getChildFolder (parentID, name) {
			try {
				for (var i = 0; ; i++) {
					var childFolderID = this.bmsvc.getIdForItemAt(parentID, i);
					var title = this.bmsvc.getItemTitle(childFolderID);
					if (name == title && this.bmsvc.getItemType(childFolderID) == this.bmsvc.TYPE_FOLDER)
						break;
				}
			}
			catch (e) {	//Rich to the end but folder not found; Create it!
				childFolderID = this.bmsvc.createFolder(parentID, name, this.bmsvc.DEFAULT_INDEX);
			}
			return childFolderID;
		}
		
		var rootFolderName = QuickFoxNotes.prefs.getComplexValue("FolderName", Components.interfaces.nsISupportsString).data;
		var arrFolders = rootFolderName.split("/");	//The folderName accept subfolders seperated by '/'
		var rootID = this.bmsvc[QuickFoxNotes.prefs.getCharPref("RootName")];
		
		for (var i = 0; i < arrFolders.length; i++) {
			rootID = getChildFolder.call (this, rootID, arrFolders[i]);	
		}
		return rootID;
	},
	getBookmarkIDs: function () {
		var bookmarkIDs = new Array();
		try {
			for (var i = 0; ; i++) {
				var childID = this.bmsvc.getIdForItemAt(this.rootID, i);
				if (this.bmsvc.getItemType(childID) == this.bmsvc.TYPE_BOOKMARK)	//Only add child bookmarks
					bookmarkIDs.push(childID);
			}
		}
		catch (e) {	//Reach to the end of list; return list
			return bookmarkIDs;
		}
	},
	getItemIndex: function (bookmarkID) {
		return this.bmsvc.getItemIndex (bookmarkID);
	},
	moveItem: function (bookmarkID, newLocation) {
		try {
			this.bmsvc.moveItem(bookmarkID, this.rootID, newLocation);	//Change position of bookmark in bookmark folder
		}catch (e) {QuickFoxNotes.jsDump.logStringMessage("QuickFox Notes: [e][io.moveItem][id:" + bookmarkID + "][e:" + e.message + "]");}
	},
	read: function (type, bookmarkID) {
		if (type == QuickFoxNotes.io.ioDescription) {
			var description = "";
			try {	//Sometime when there is no text in the bookmark's description, error happens
				description = this.annotationService.getItemAnnotation(bookmarkID, "bookmarkProperties/description");
			} catch (e) {QuickFoxNotes.jsDump.logStringMessage("QuickFox Notes: [e][io.read(); io.ioDescription][id:" + bookmarkID + "][e:" + e.message + "]");}
			return description;
		}
		if (type == QuickFoxNotes.io.ioTitle)
			return this.bmsvc.getItemTitle(bookmarkID);
	},
	write: function (type, bookmarkID, data) {	//if bookmarkID == null and type == ioTitle; then new bookmark will be created
		if (type == QuickFoxNotes.io.ioDescription) {
			this.annotationService.setItemAnnotation(bookmarkID, "bookmarkProperties/description", data, 0, this.annotationService.EXPIRE_NEVER);
		}
		if (type == QuickFoxNotes.io.ioTitle) {
			if (bookmarkID)
				this.bmsvc.setItemTitle(bookmarkID, data);
			else {	//Create new bookmark
				var ios = Components.classes["@mozilla.org/network/io-service;1"]
					.getService(Components.interfaces.nsIIOService);
				var uri = ios.newURI("chrome://qfnotes/content/openEditor.xul?random=" + Math.floor( Math.random() * 1000001 ), null, null);	//Prevent duplicate bookmarks
				return this.bmsvc.insertBookmark(this.rootID, uri, this.bmsvc.DEFAULT_INDEX, data);
			}
		}
		return null;
	},
	remove: function (arrBookmarkIDs) {
		for (var i = 0; i < arrBookmarkIDs.length; i++) {
			try {
				this.bmsvc.removeItem(arrBookmarkIDs[i]);
			} catch(e) {QuickFoxNotes.jsDump.logStringMessage("QuickFox Notes: [e][io.remove()][e:" + e.message + "]");}
		}
		return this;	
	}
};
/**Local file I/O service; Combined mode*/
QuickFoxNotes.io.sqlite = {	//This script is modified version of http://codesnippets.joyent.com/posts/show/1030#related
	/**Internal functions*/
	storageService: null,
	mDBConn: null,
	_initService: function() {
		var path = QuickFoxNotes.prefs.getComplexValue("sqlitePath", Components.interfaces.nsISupportsString).data;
	
		var db = Components.classes["@mozilla.org/file/local;1"].createInstance(Components.interfaces.nsILocalFile);  
		db.initWithPath(path);
		this.storageService = Components.classes["@mozilla.org/storage/service;1"].getService(Components.interfaces.mozIStorageService);
		this.mDBConn = (this.storageService).openDatabase(db);
		//Create Database if it doesnt exist
		this._cmd('CREATE TABLE IF NOT EXISTS quickfox (id INTEGER PRIMARY KEY AUTOINCREMENT, iOrder INTEGER, title TEXT, body TEXT);');	
	},
	_select: function(sql,param) {
		if (this.storageService== undefined){
                    this._initService();
		}
		var ourTransaction = false;
		if ((this.mDBConn).transactionInProgress){
			ourTransaction = true;
			(this.mDBConn).beginTransactionAs((this.mDBConn).TRANSACTION_DEFERRED);
		}
		var statement = (this.mDBConn).createStatement(sql);
                if (param){
			for (var m=2, arg=null; arg=arguments[m]; m++) {
				statement.bindUTF8StringParameter(m-2, arg);
			}
		}
		try{
			var dataset = [];
			while (statement.executeStep()){
				var row = [];
				for(var i=0,k=statement.columnCount; i<k; i++){
					row[statement.getColumnName(i)] = statement.getUTF8String(i);
				}
				dataset.push(row);
			}
			// return dataset;	
		}
		finally {
			statement.reset();
		}
		if (ourTransaction){
			(this.mDBConn).commitTransaction();
		}
        return dataset;	
	},
	_cmd: function(sql,param) {
		if (this.storageService == undefined){
	                    this._initService();
		}
		var ourTransaction = false;
		if ((this.mDBConn).transactionInProgress){
			ourTransaction = true;
			(this.mDBConn).beginTransactionAs((this.mDBConn).TRANSACTION_DEFERRED);
		}
		var statement = (this.mDBConn).createStatement(sql);
		if (param){
			for (var m=2, arg=null; arg=arguments[m]; m++) {
				statement.bindUTF8StringParameter(m-2, arg);
			}
		}
		try{
			statement.execute();
		}
		finally {
			statement.reset();
		}
		if (ourTransaction){
			(this.mDBConn).commitTransaction();
		}
	},
	/**End of Internal functions*/
	/**Initialize SQLITE data storage*/
	initialize: function () {
		var path = QuickFoxNotes.prefs.getComplexValue("sqlitePath", Components.interfaces.nsISupportsString).data;
		if (!path) {
			var nsIFilePicker = Components.interfaces.nsIFilePicker;
			var fp = Components.classes["@mozilla.org/filepicker;1"].createInstance(nsIFilePicker);
			fp.init(window, QuickFoxNotes.stringsBundle.getString('io1'), nsIFilePicker.modeGetFolder);
			var res = fp.show();
			if (res == nsIFilePicker.returnOK){
				// "\\" for windows and "/" for linux and mac
				var pathDivider = (Components.classes["@mozilla.org/xre/app-info;1"].getService(Components.interfaces.nsIXULRuntime).OS == "WINNT") ? "\\" : "/";  
				path = fp.file.path + pathDivider + "qfn.sqlite";

				var str = Components.classes["@mozilla.org/supports-string;1"]
					.createInstance(Components.interfaces.nsISupportsString);
				str.data = path;
				QuickFoxNotes.prefs.setComplexValue("sqlitePath", Components.interfaces.nsISupportsString, str); 
			}
			else {
				var userAnswer = 1;
				if (!window.confirm(QuickFoxNotes.stringsBundle.getString('io2'))) {
					QuickFoxNotes.prefs.setIntPref("repositoryType", 0);
					userAnswer = 0;
				}
				
				return (userAnswer ? -1 : false);	//This is only used in ioSwitch
			}		
		}
		return true;
	},
	releaseDatabase: function () {
		if (this.mDBConn.mozIStorageConnection.asyncClose) {
			this.mDBConn.mozIStorageConnection.asyncClose();
			this.storageService = null;
		}
	},
	/**Get free index for location; new item located in thje end of list*/
	_freeIndex: function () {
		try {
			var arrData = this._select('SELECT iOrder FROM quickfox ORDER BY iOrder');
			return parseInt(arrData[arrData.length - 1]['iOrder']) + 1;	//String instead of number
		}
		catch(e) {	//Return one on empty repository
			return 1;
		}	
	},
	/**Get All available indices*/
	getIndices: function () {
		try {
			var arrData = this._select('SELECT id FROM quickfox ORDER BY iOrder');
		}
		catch (e) {	//Return empty on empty repository
			return [];
		}
		var indices = []
		for (var i = 0; i < arrData.length; i++) {
			indices[i] = parseInt(arrData[i]['id']);
		}
		return indices;
	},
	/**Read*/
	read: function (type, ID) {
		var key;
		if (type == QuickFoxNotes.io.ioDescription)
			key = "body"
		else
			key = "title"

		var temp = this._select('SELECT ' + key + ' FROM quickfox WHERE id=' + ID + ';');
		return temp[0][key].replace (/&39;/g, "'").replace (/&34;/g, '"');
	},
	/**Write*/
	write: function (type, ID, data) {	//if ID == null and type == ioTitle; then new bookmark will be created
		data = data.replace (/\'/g, "&39;").replace (/\"/g, "&34;");	
	
		if (type == QuickFoxNotes.io.ioDescription) {
			this._cmd('UPDATE quickfox SET body="' + data + '" WHERE id=' + ID + ';');
		}
		if (type == QuickFoxNotes.io.ioTitle) {
			if (ID)
				this._cmd('UPDATE quickfox SET title="' + data + '" WHERE id=' + ID + ';');
			else {	//Add new Note
				var newIndex = this._freeIndex();
				this._cmd('INSERT INTO quickfox(iOrder, title, body) VALUES(' + newIndex + ', "' + data + '", "");');
				return parseInt(this._select('SELECT id FROM quickfox WHERE iOrder=' + newIndex + ';')[0]['id']);
			}
		}
		return null;	
	},
	/**Get item index*/
	getItemIndex: function (ID) {
		try {
			var temp = this._select('SELECT iOrder FROM quickfox WHERE id=' + ID + ';');
			return parseInt(temp[0]['iOrder']);
		}
		catch(e) {
			return -1;
		}
	},
	/**Change index of item to locate into new location*/
	moveItem: function (ID, newLocation) {
		var oldLocation = this.getItemIndex(ID);		
		
		if (oldLocation < newLocation) {
			newLocation -= 1;	//Who knows why!
			this._cmd('UPDATE quickfox SET iOrder = iOrder - 1 WHERE iOrder >= ' + (oldLocation + 1) + ' AND iOrder <= ' + newLocation + ';');
		}
		else {
			this._cmd('UPDATE quickfox SET iOrder = iOrder + 1 WHERE iOrder >= ' + newLocation + ' AND iOrder <= ' + (oldLocation - 1) + ';');
		}
		this._cmd('UPDATE quickfox SET iOrder = ' + newLocation + ' WHERE id = ' + ID + ';');
	},
	/**Remove series of Notes*/
	remove: function (IDs) {
		if (!IDs.length)
			return;
		var data = "id = " + IDs[0];
		for (var i = 1; i < IDs.length; i++) {
			data += " OR id = " + IDs[i]
		}
		this._cmd('DELETE FROM quickfox WHERE ' + data);
		
		return this;
	}
};
/**Local file I/O service; Seperate mode*/
QuickFoxNotes.io.local = {
	rFile: null, //Root directory

	files: [],
	
	nameToBeIdentifier: String.fromCharCode(0) + String.fromCharCode(2) + String.fromCharCode(1) + String.fromCharCode(0),
	nameToBe: [],	//This array contains the modified name in case there would be rename problem 
	
	initialize: function () {
		this.rFile = Components.classes["@mozilla.org/file/local;1"].  
                   createInstance(Components.interfaces.nsILocalFile);
		var path = QuickFoxNotes.prefs.getComplexValue("lFolderName", Components.interfaces.nsISupportsString).data;
		if (!path) {
			var nsIFilePicker = Components.interfaces.nsIFilePicker;
			var fp = Components.classes["@mozilla.org/filepicker;1"].createInstance(nsIFilePicker);
			fp.init(window, QuickFoxNotes.stringsBundle.getString('io1'), nsIFilePicker.modeGetFolder);
			var res = fp.show();
			if (res == nsIFilePicker.returnOK){
				path = fp.file.path;
				//
				var str = Components.classes["@mozilla.org/supports-string;1"]
					.createInstance(Components.interfaces.nsISupportsString);
				str.data = path;
				QuickFoxNotes.prefs.setComplexValue("lFolderName", Components.interfaces.nsISupportsString, str); 
			}
			else {
				var userAnswer = 1;
				if (!window.confirm(QuickFoxNotes.stringsBundle.getString('io2'))) {
					QuickFoxNotes.prefs.setIntPref("repositoryType", 0);
					userAnswer = 0;
				}
				
				return (userAnswer ? -1 : false);	//This is only used in ioSwitch
			}		
		}
		this.rFile.initWithPath(path);
		
		return true;
	},
	format: {
		toQFN: function (data) {
			return data.replace (/&92;/g, "\\")
					   .replace (/&47;/g, "/")
					   .replace (/&58;/g, ":")
					   .replace (/&42;/g, "*")
					   .replace (/&63;/g, "?")
					   .replace (/&34;/g, "\"")
					   .replace (/&62;/g, ">")
					   .replace (/&60;/g, "<")
					   .replace (/&124;/g, "|");
		},
		toFile: function (data) {
			return data.replace (/\\/g, "&92;")
					   .replace (/\//g, "&47;")
					   .replace (/\:/g, "&58;")
					   .replace (/\*/g, "&42;")
					   .replace (/\?/g, "&63;")
					   .replace (/\"/g, "&34;")
					   .replace (/\>/g, "&62;")
					   .replace (/\</g, "&60;")
					   .replace (/\|/g, "&124;");
		}
	},
	rename: function (index, newName) {
		var file = this.files[index];
		if (!file)
			return;
			
		var content = this.read (QuickFoxNotes.io.ioDescription, index);
		
		try {	//Try to rename the note; if any error occurs then save the changes in the body part; getName() will take care of rename proccess later
			file.moveTo(null, this.format.toFile(newName));
			this.nameToBe[index] = null;
			//Repair content
			this.write(QuickFoxNotes.io.ioDescription, index, content);
		} catch (e) {
			this.nameToBe[index] = newName;
			//Overwrite modify section in content
			this.write(QuickFoxNotes.io.ioDescription, index, content + this.nameToBeIdentifier + newName);
		}
	},
	getName: function(index) {
		//In case there was a rename problem	
		var modifications = this.read (QuickFoxNotes.io.ioDescription, index, true).split(this.nameToBeIdentifier)[1];
		if (modifications)	//Try to rename the note
			this.rename(index, modifications);//Try to rename the note		
		
		var entireName;
		if (this.nameToBe[index]) 
			entireName = this.nameToBe[index];
		else {
			var file = this.files[index];
			entireName = file.leafName;
		}

		var array = entireName.match(/^#\d+\ (.*)/);
		if (array)
			return this.format.toQFN(array[1]);
		return this.format.toQFN(entireName);
	},
	getFilesIndex: function () {
		var entries = this.rFile.directoryEntries;
		var noIndex = [];

		while(entries.hasMoreElements()) {
			var entry = entries.getNext();
			entry.QueryInterface(Components.interfaces.nsIFile);
			
			
			var index = this.getItemIndex(entry);
			if (index != -1)	//If two files have same index, re-index one of them
				this.files[index] = entry
			else
				noIndex.push(entry);
		}
		//Repair no index text files
		for (var i = 0; i < noIndex.length; i++)
			for (j = 0; ; j++)
				if (!this.files [j]) {
					this.files [j] = noIndex[i];
					this.rename(j, "#" + j + " " + noIndex[i].leafName);
					break;
				}
		
		return function(me) {
			var temp =[];
			for (i = 0; i < me.files.length; i++) {
				if (me.files[i])
					temp.push(i);
			}
			return temp;
		}(this);
	},
	getItemIndex: function (index) {
		var file;
		if (typeof index == "number")		//Only for intitialize process index is a file obj
			file = this.files[index];
		else
			file = index;
		
		const intirename = file.leafName;
		var array = intirename.match(/^#(\d+)\ /);
		
		if (array)
			return parseInt(array[1]);
		return -1;
	},
	moveItem: function (index, newLocation) {
		var file = this.files[index];
		var oldIndex = this.getItemIndex(index);
		var tempFile = file;
		
		for (var i = oldIndex + 1; i <= newLocation; i++) {
			this.files[i - 1] = this.files[i];
			if (this.files[i])
				this.rename(i - 1, "#" + (i - 1) + " " + this.getName(i - 1));
		}
		
		for (var i = oldIndex - 1; i >= newLocation; i--) {
			this.files[i + 1] = this.files[i];
			if (this.files[i])
				this.rename(i + 1, "#" + (i + 1) + " " + this.getName(i + 1));
		}
		this.files[newLocation] = tempFile;
		this.rename(newLocation, "#" + (newLocation) + " " + this.getName(newLocation));
	},
	read: function (type, index, bareText) {
		var file = this.files[index];
		
		if (type == QuickFoxNotes.io.ioTitle)	//Convert name to valid label
			return this.format.toQFN(this.getName(index));
		if (type == QuickFoxNotes.io.ioDescription) {		
			var data = "";
			var fstream = Components.classes["@mozilla.org/network/file-input-stream;1"].
									createInstance(Components.interfaces.nsIFileInputStream);
			var cstream = Components.classes["@mozilla.org/intl/converter-input-stream;1"].
									createInstance(Components.interfaces.nsIConverterInputStream);
			fstream.init(file, -1, 0, 0);
			cstream.init(fstream, "UTF-8", 0, 0); // you can use another encoding here if you wish


			var str = {};
			var rd = 0;
			do {
				rd = cstream.readString(0xffffffff, str); // read as much as we can and put it in str.value
				data += str.value;
			} while (rd != 0);

			cstream.close(); // this closes fstream
		
			if (bareText)
				return data;
			return data.split(this.nameToBeIdentifier)[0];	//In case if there were title modifications in the content
		}
	},
	write: function (type, index, data) {	//if file == null and type == ioTitle; then new file will be created
		var file = this.files[index];
	
		if (type == QuickFoxNotes.io.ioTitle) {
			//Convert name to valid label
			data = this.format.toFile(data.replace (/\\/g, "&92;"));
	
			if (file)
				this.rename(index, "#" + this.getItemIndex(index) + " " + data);
			else {	//Create new file
				// "\\" for windows and "/" for linux and mac
				var pathDivider = (Components.classes["@mozilla.org/xre/app-info;1"].getService(Components.interfaces.nsIXULRuntime).OS == "WINNT") ? "\\" : "/";  
				var nfile = Components.classes["@mozilla.org/file/local;1"]
					.createInstance(Components.interfaces.nsILocalFile); 
				nfile.initWithPath(
					QuickFoxNotes.prefs.getComplexValue("lFolderName", Components.interfaces.nsISupportsString).data + 
					pathDivider + "#" + this.files.length + " " + data);
				var foStream = Components.classes["@mozilla.org/network/file-output-stream;1"]
					.createInstance(Components.interfaces.nsIFileOutputStream);
				foStream.init(nfile, 0x02 | 0x08 | 0x20, 0666, 0);
				
				this.files[this.files.length] = nfile;
				
				return (this.files.length - 1);
			}
		}
		if (type == QuickFoxNotes.io.ioDescription) {
			var foStream = Components.classes["@mozilla.org/network/file-output-stream;1"].
									 createInstance(Components.interfaces.nsIFileOutputStream);

			// use 0x02 | 0x10 to open file for appending.
			foStream.init(file, 0x02 | 0x08 | 0x20, 0666, 0); 

			var converter = Components.classes["@mozilla.org/intl/converter-output-stream;1"].
									  createInstance(Components.interfaces.nsIConverterOutputStream);
			converter.init(foStream, "UTF-8", 0, 0);
			//Check if there is any title modifications
			converter.writeString(this.nameToBe[index] ? data + this.nameToBeIdentifier + this.nameToBe[index] : data);
			converter.close(); // this closes foStream
		}
		return null;
	},
	remove: function (indices) {
		for (var i = 0; i < indices.length; i++) {
			var file = this.files[indices[i]];
			try {
				file.remove(false);
			} catch(e) {
				alert('QFN could not remove the file!');
			}
		}
		return this;
	}
};